//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Class          : BigRational
//
// Author         : Richard E. Pattis
//                  Computer Science Department
//                  Carnegie Mellon University
//                  5000 Forbes Avenue
//                  Pittsburgh, PA 15213-3891
//                  e-mail: pattis@cs.cmu.edu
//
// Maintainer     : Author
//
//
// Description:
//
//   Objects constructed from the BigRational class act like fractions
// (rational numbers have a sign, numerator, and denominator). This
// class implements some standard arithmetic and relational operations
// (as methods) on values of this class.
//
//
// Known Bugs (if any): None
//
// Future Plans       : Document via JavaDoc (Done: 8/10/04)
//
// Program History:
//   9/18/01: R. Pattis - Operational for 15-100
//   8/10/04: R. Pattis - Javadoc comments added (added to course library)
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


package edu.cmu.cs.pattis.cs151xx;


import java.math.BigInteger;



/** 
 * Objects constructed from the BigRational class act like fractions
 *   (rational numbers have a sign, numerator, and denominator).
 * This class implements some standard arithemetic and relational
 *  operations (as methods) on values of this class.
*/
public class BigRational {


	/** 
	 * Constructs a <code>BigRational</code> object, specifying the
	 *   numerator and denominator of the fraction.
	 * These value are examined and put in a canonical form to ensure 
	 *   the following class invariants
	 *   <ul>
	 *     <li>Zero is always stored as 0/1
	 *     <li>Denominators are always positive
	 *     <li>Numerators and Denominators are in lowest common terms
	 *           (they have no common factors)
	 *   </ul>
	 *
	 * @param  numerator specifies the numerator of the fraction
	 * @param  denominator specifies the denominator of the fraction
	 *
	 * @throws IllegalArgumentException if <code>denominator</b> is 0
	*/
  public BigRational (BigInteger numerator, BigInteger denominator)
    throws IllegalArgumentException
  {
    if (denominator.equals(BigInteger.ZERO))
      throw new IllegalArgumentException("Exception in BigRational Construtor; - denominator 0");

    //Ensure 0 is stored as 0/1
    if (numerator.equals(BigInteger.ZERO))
      denominator = BigInteger.ONE;
    
    //Ensure non-negative denominator; if a fraction is negative, its
    //  numerator is negative
    if (denominator.compareTo(BigInteger.ZERO) < 1) {
      denominator = denominator.negate();
      numerator   = numerator.negate();
    }
    
    BigInteger common = numerator.gcd(denominator);
        
    //Ensure that there are no common factors
    this.numerator   = numerator.divide(common);
    this.denominator = denominator.divide(common);
  }


  
	/** 
	 * Constructs a <code>BigRational</code> object, specifying the
	 *   numerator of the fraction (the denominator is one).
	 *
	 * @param  numerator specifies the numerator of the fraction
	*/
  public BigRational (BigInteger numerator)
  {this(numerator, BigInteger.ONE);}
  
  
  
	/** 
	 * Constructs a <code>BigRational</code> object 0/1.
	 *
	 * @param  numerator specifies the numerator of the fraction
	*/
  public BigRational ()
  {this(BigInteger.ZERO, BigInteger.ONE);}
  
  
  
	/** 
	 * Constructs a <code>BigRational</code> object whose state is the
	 *   same as <code>other</code>.
	 *
	 * @param other specifies the fraction whose value to store into
	 *          the state of this object
	*/
  public BigRational (BigRational other)
  {this(other.numerator, other.denominator);}
  
  
  
  
	/** 
	 * Returns the numerator of the fraction.
	 *
	 * @return the numerator of the fraction
	*/
  public BigInteger getNumerator()
  {return this.numerator;}
  
  
  
	/** 
	 * Returns the denominator of the fraction.
	 *
	 * @return the denominator of the fraction
	*/
  public BigInteger getDenominator()
  {return this.denominator;}
  
  
  
	/** 
	 * Returns whether or not this fraction is equal to <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the equality check
	 *
	 * @return whether or not this fraction is equal to <code>other</code>
	*/
  public boolean equals (BigRational other)
  {return this.numerator.equals(other.numerator) && this.denominator.equals(other.denominator);}
  

  
	/** 
	 * Returns the trichotomous comparison of this value and
	 *   <code>other</code>.
	 * More specifically:
	 * <ul>
	 *   <li>a negative value if <code>this</code> &lt <code>other</code>
	 *   <li>zero if <code>this</code> = <code>other</code>
	 *   <li>a positive value if <code>this</code> &gt <code>other</code>
	 * </ul>
	 * Example use (checing whether <code>f1</code> is less than or equal
	 *   to <code>f2</code>:<p>
	 *   <b><pre>  if (f1.compareTo(f2) <= 0)</pre></b>
	 *
	 * @param  other specifies the other fraction in the comparison
	 *
	 * @return the trichotomous comparison of this value and
	 *   <code>other</code>
	*/
  public int compareTo (BigRational other)
  {
    BigInteger left  = this.numerator.multiply(other.denominator);
    BigInteger right = this.denominator.multiply(other.numerator);
    
    //  a     c
    // --- ? --- = ad ? bc
    //  b     d
    
    return left.compareTo(right);
  }
  
  
	/** 
	 * Returns a new fraction that is the absolute value of this one.
	 *
	 * @return a new fraction that is the absolute value of this one.
	*/
  public BigRational abs()
  {return new BigRational(this.numerator.abs(),this.denominator);}
  
  
  
	/** 
	 * Returns a new fraction that is the negated value of this one.
	 *
	 * @return a new fraction that is the negated value of this one.
	*/
  public BigRational negate()
  {return new BigRational(this.numerator.negate(),this.denominator);}
  
  
  
	/** 
	 * Returns a new fraction that is the reciprocal value of this one:
	 *   1/<code>this</code>.
	 *
	 * @return a new fraction that is the reciprocal value of this one:
	 *   1/<code>this</code>
	 *
	 * @throws IllegalStateException if this value is 0
	*/
  public BigRational reciprocal()
    throws IllegalStateException
  {
    //In this version, check directly (see divide for another way)
    if (this.equals(ZERO))
      throw new IllegalStateException("Exception in BigRational reciprocal: attempt to invert zero");

    return new BigRational(this.denominator,this.numerator);
  }
  
  
  
	/** 
	 * Returns a new fraction that is the sum of this one and
	 *   <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the sum
   *
	 * @return a new fraction that is the sum of this one and
	 *   <code>other</code>
	*/
  public BigRational add (BigRational other)
  { 
    BigInteger a = this.numerator;          //  a     c     ad + cb
    BigInteger b = this.denominator;        // --- + --- = ---------
    BigInteger c = other.numerator;         //  b     d        bd
    BigInteger d = other.denominator;       //
    
    return new BigRational(a.multiply(d).add(c.multiply(b)), b.multiply(d));
  }
  


	/** 
	 * Returns a new fraction that is the difference of this one and
	 *   <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the difference
   *
	 * @return a new fraction that is the difference of this one and
	 *   <code>other</code>
	*/
  public BigRational subtract (BigRational other)
  { 
    BigInteger a = this.numerator;          //  a     c     ad - cb
    BigInteger b = this.denominator;        // --- - --- = ---------
    BigInteger c = other.numerator;         //  b     d        bd
    BigInteger d = other.denominator;       //
    
    return new BigRational(a.multiply(d).subtract(c.multiply(b)), b.multiply(d));
    
    //or write the following, which is compact but a bit slower
    //  this.add(other.negate())
  }
  
  
  
	/** 
	 * Returns a new fraction that is the product of this one and
	 *   <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the product
   *
	 * @return a new fraction that is the product of this one and
	 *   <code>other</code>
	*/
  public BigRational multiply (BigRational other)
  { 
    BigInteger a = this.numerator;          //  a     c     ac
    BigInteger b = this.denominator;        // --- * --- = ----
    BigInteger c = other.numerator;         //  b     d     bd
    BigInteger d = other.denominator;       //
    
    return new BigRational(a.multiply(c),b.multiply(d));
  }
  


	/** 
	 * Returns a new fraction that is the dividend of this one and
	 *   <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the division
   *
	 * @return a new fraction that is the dividend of this one and
	 *   <code>other</code>
	 *
	 * @throws IllegalStateException if <code>other</b> value is 0
	*/
  public BigRational divide (BigRational other)
    throws IllegalArgumentException
  { 
    BigInteger a = this.numerator;          //  a     c     ad
    BigInteger b = this.denominator;        // --- / --- = ----
    BigInteger c = other.numerator;         //  b     d     bc
    BigInteger d = other.denominator;       //
    
    return new BigRational(a.multiply(d),b.multiply(c));

    //or write the following, which is compact but a bit slower
    //  this.multiply(other.reciprocal())
  }
  


	/** 
	 * Returns a <code>String<code> representation of the state of this 
	 *  fraction: numerator/denominator (which can be unreadable if the
	 *  numerator and denominator are large).
	 *
	 * @return a <code>String<code> representation of the state of this 
	 *  fraction: numerator/denominator
	*/
  public String toString()
  {return this.numerator + "/" + this.denominator;}
  
  

	/** 
	 * Returns a <code>String<code> representation of the state of this 
	 *  fraction, as a decimal value approximating the fraction.
	 * For example <code>(new BigRational(""+1,""+3)).toDecimalString(20)</code>
	 *   returns <code>".33333333333333333333"</code>;
	 *  <code>(new Rational(""+1,""+7)).toDecimalString(13)</code>
	 *   returns <code>".1428571428571"</code>.
	 *
	 * @param  decimalPlaces specifies the number of decimal places to use
	 *           in the approximation
   *
	 * @return a <code>String<code> representation of the state of this 
	 *  fraction, as a decimal value approximating the fraction
	*/
  public String toDecimalString(int decimalPlaces)
  {
    String     answer = new String("");
    BigInteger ten    = new BigInteger("10");
    
    BigInteger num   = this.numerator;
    BigInteger denom = this.denominator;
    
    //Handle negative numbers
    if (num.compareTo(BigInteger.ZERO) < 0) {
      num = num.negate();
      answer = answer + "-";
    }
    
    //Handle integer part
    if (num.compareTo(denom) >= 0) {
      answer += num.divide(denom);
      num     = num.subtract(num.divide(denom).multiply(denom));
    }

    //Handle decimal part
    answer = answer + ".";
    num = num.multiply(ten);
    for (int i=1; i<=decimalPlaces; i++) {
	    answer += num.divide(denom);
	    num    = num.subtract(num.divide(denom).multiply(denom));
      num    = num.multiply(ten);
	  }
      
    return answer;
  }
  
  
  
	/** 
	 * Returns a new fraction after prompting the user for its numerator
	 *   and denominator.
	 * If an illegal value is entered, the user is reprompted
	 * For example, if we write <code>BigRational.prompt("Enter height")</code>
	 *   an interaction on the console screen might look like
	 * <b><pre>  Enter Height
	 *    Enter numerator  : 1
	 *    Enter denominator: 3</pre></b>
	 *
	 * @param message specifies the message with which to prompt the user
   *
	 * @return a new fraction after prompting the user for its numerator
	 *   and denominator.
	*/
  public static BigRational prompt(String s)
  {
    System.out.println(s);
    for (;;)
      try{
        BigInteger numerator   = Prompt.forBigInteger("  Enter numerator  ");
        BigInteger denominator = Prompt.forBigInteger("  Enter denominator");
        return new BigRational(numerator,denominator);
      }catch (IllegalArgumentException iae)
        {System.out.println("Illegal BigRational entered; please try again");}
  }






	/** 
	 * Stores the numerator of the fraction.
	 * It may be positive or negative.
	 * It will have no common factors with the denominator.
	*/
  private BigInteger numerator;
  
  
	/** 
	 * Stores the denominator of the fraction.
	 * It is always positive, and stores 1 if the numerator is 0.
	 * It will have no common factors with the numerator.
	*/
  private BigInteger denominator;
  
  
	/** 
	 * Stores the fraction 0/1.
	*/
  public static final BigRational ZERO = new BigRational();
  
  
  /** 
	 * Stores the fraction 1/1.
	*/
  public static final BigRational ONE  = new BigRational(BigInteger.ONE);
  
  
}
